"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createDepSymbolResolver = exports.isInTypeContext = void 0;
const constants_1 = require("./constants");
const helper_1 = require("./helper");
const utils_1 = require("./utils");
function isInTypeContext(typescript, node) {
    return typescript.isTypeElement(node) || typescript.isPartOfTypeNode(node);
}
exports.isInTypeContext = isInTypeContext;
function createDepSymbolResolver(typescript, scope, additionalScope, file, checker, preferConstantsCall, logger) {
    const cached = new Map();
    const duplicated = new Set();
    const accessExpressionContainsInnerCached = new Map();
    return {
        shouldSymbolDefinitelyBeIgnoreInDeps,
        alreadyDuplicated,
        markAsDuplicated,
        isExpressionContainsDeclaredInner,
        markExpressionContainsDeclaredInner
    };
    function isWellKnownGlobalSymbol(symbol) {
        return (checker.isUndefinedSymbol(symbol) ||
            checker.isArgumentsSymbol(symbol));
    }
    function markExpressionContainsDeclaredInner(expr, value) {
        if (accessExpressionContainsInnerCached.has(expr)) {
            return;
        }
        accessExpressionContainsInnerCached.set(expr, value);
    }
    function isDeclarationContainsInnerFile(valueDeclaration) {
        if (valueDeclaration.getSourceFile() !== file) {
            return false;
        }
        if (typescript.rangeContainsRange(scope, valueDeclaration)) {
            return true;
        }
        if (additionalScope.some(s => typescript.rangeContainsRange(s, valueDeclaration))) {
            return true;
        }
        return false;
    }
    function isExpressionContainsDeclaredInner(expr) {
        return visitor(expr);
        function visitor(node) {
            if (!accessExpressionContainsInnerCached.has(node)) {
                if (typescript.isTypeNode(node)) {
                    return undefined;
                }
                if (typescript.isIdentifier(node)) {
                    const rawSymbol = checker.getSymbolAtLocation(node);
                    const symbol = rawSymbol &&
                        utils_1.skipSymbolAlias(typescript, rawSymbol, checker);
                    if (symbol &&
                        !isWellKnownGlobalSymbol(symbol) &&
                        symbol.valueDeclaration) {
                        const result = isDeclarationContainsInnerFile(symbol.valueDeclaration);
                        accessExpressionContainsInnerCached.set(node, result);
                        return result;
                    }
                    accessExpressionContainsInnerCached.set(node, false);
                    return false;
                }
                const result = typescript.forEachChild(node, visitor);
                accessExpressionContainsInnerCached.set(node, !!result);
                return result;
            }
            return accessExpressionContainsInnerCached.get(node);
        }
    }
    function peekSymbolDefinitelyBeIgnoreInDeps(rawSymbol) {
        if (isWellKnownGlobalSymbol(rawSymbol)) {
            return true;
        }
        const symbol = utils_1.skipSymbolAlias(typescript, rawSymbol, checker);
        const valueDeclaration = symbol.valueDeclaration;
        if (!valueDeclaration) {
            return false;
        }
        logger.log('WTF 3');
        const declFile = valueDeclaration.getSourceFile();
        if (declFile.isDeclarationFile) {
            return true;
        }
        const inTypeContext = isInTypeContext(typescript, valueDeclaration);
        if (inTypeContext) {
            return false;
        }
        logger.log('WTF 4');
        if (declFile !== file) {
            return true;
        }
        if (isDeclarationContainsInnerFile(valueDeclaration)) {
            return true;
        }
        logger.log('WTF 5');
        if (utils_1.isDeclarationAssignedByUseRef(typescript, valueDeclaration) ||
            utils_1.isDeclarationAssignedByUseState(typescript, valueDeclaration) ||
            utils_1.isDeclarationAssignedByUseReducer(typescript, valueDeclaration) ||
            utils_1.isDeclarationAssignedByConstantsUseMemo(typescript, valueDeclaration) ||
            utils_1.isDeclarationAssignedByConstantsUseCallback(typescript, valueDeclaration)) {
            return true;
        }
        logger.log('WTF 6');
        if (isDeclarationConstants(valueDeclaration)) {
            return true;
        }
        logger.log('WTF 7');
        if (utils_1.isTopLevelConstantDeclaration(typescript, valueDeclaration)) {
            return true;
        }
        logger.log('WTF 8');
        return false;
        function isExpressionConstants(node) {
            if (typescript.isLiteralExpression(node)) {
                return true;
            }
            if (typescript.isParenthesizedExpression(node) ||
                typescript.isNonNullExpression(node)) {
                return isParenExpressionOrNonNullExpressionConstants(node);
            }
            if (typescript.isIdentifier(node)) {
                return isIdentifierConstants(node);
            }
            if (typescript.isPropertyAccessExpression(node)) {
                return isPropertyAccessConstants(node);
            }
            if (typescript.isElementAccessExpression(node)) {
                return isElementAccessConstants(node);
            }
            if (typescript.isTemplateExpression(node)) {
                return isTemplateLiteralConstants(node);
            }
            if (typescript.isBinaryExpression(node)) {
                return isBinaryExpressionConstants(node);
            }
            if (typescript.isPrefixUnaryExpression(node) ||
                typescript.isPostfixUnaryExpression(node)) {
                return isUpdateExpressionConstant(node);
            }
            if (typescript.isCallExpression(node)) {
                return isCallExpressionConstants(node);
            }
            if (utils_1.isFunctionExpressionLike(typescript, node)) {
                return isSimpleFunctionLikeConstants(node);
            }
            return false;
        }
        function isDeclarationConstants(decl) {
            if (typescript.isEnumDeclaration(decl) ||
                typescript.isEnumMember(decl) ||
                typescript.isModuleDeclaration(decl)) {
                return true;
            }
            if (typescript.isVariableDeclaration(decl) &&
                typescript.getCombinedNodeFlags(decl) &
                    typescript.NodeFlags.Const &&
                decl.initializer &&
                isExpressionConstants(decl.initializer)) {
                return true;
            }
            if (utils_1.isFunctionExpressionLike(typescript, decl)) {
                return isSimpleFunctionLikeConstants(decl);
            }
            return false;
        }
        function isParenExpressionOrNonNullExpressionConstants(expression) {
            return isExpressionConstants(expression.expression);
        }
        function isSimpleFunctionLikeConstants(func) {
            if (func.parameters.length !== 0) {
                return false;
            }
            if (typescript.isBlock(func.body)) {
                const effectiveStatement = func.body.statements.filter(stmt => typescript.isEmptyStatement(stmt));
                if (effectiveStatement.length === 1) {
                    const firstStatement = helper_1.first(effectiveStatement);
                    if (typescript.isReturnStatement(firstStatement) &&
                        firstStatement.expression) {
                        return isExpressionConstants(firstStatement.expression);
                    }
                }
            }
            else {
                return isExpressionConstants(func.body);
            }
            return false;
        }
        function isCallExpressionConstants(call) {
            logger.log('isCallExpressionConstants: ' + call.getText(file));
            logger.log('isCallExpressionConstants: ' + preferConstantsCall);
            if (!preferConstantsCall) {
                return false;
            }
            logger.log('isCallExpressionConstants: ' + 'check arguments');
            if (!call.arguments.every(isExpressionConstants)) {
                return false;
            }
            logger.log('isCallExpressionConstants: ' + 'check expression');
            return isExpressionConstants(call.expression);
        }
        function isBinaryExpressionConstants(expression) {
            if (typescript.isAssignmentExpression(expression)) {
                return false;
            }
            return (isExpressionConstants(expression.left) &&
                isExpressionConstants(expression.right));
        }
        function isUpdateExpressionConstant(expression) {
            if (expression.operator === typescript.SyntaxKind.PlusPlusToken ||
                expression.operator === typescript.SyntaxKind.MinusMinusToken) {
                return false;
            }
            return isExpressionConstants(expression.operand);
        }
        function isIdentifierImmutable(identifier) {
            return (identifier.text === constants_1.Constants.UndefinedKeyword &&
                identifier.originalKeywordKind ===
                    typescript.SyntaxKind.UndefinedKeyword);
        }
        function isIdentifierConstants(expression) {
            if (isIdentifierImmutable(expression)) {
                return true;
            }
            const symbol = checker.getSymbolAtLocation(expression);
            return !!(symbol && shouldSymbolDefinitelyBeIgnoreInDeps(symbol));
        }
        function isLiteralImmutable(expression) {
            switch (expression.kind) {
                case typescript.SyntaxKind.StringLiteral:
                case typescript.SyntaxKind.NoSubstitutionTemplateLiteral:
                case typescript.SyntaxKind.NumericLiteral:
                case typescript.SyntaxKind.TrueKeyword:
                case typescript.SyntaxKind.FalseKeyword:
                case typescript.SyntaxKind.NullKeyword:
                case typescript.SyntaxKind.RegularExpressionLiteral:
                    return true;
                case typescript.SyntaxKind.Identifier:
                    return isIdentifierImmutable(expression);
                default:
                    return false;
            }
        }
        function isPropertyAccessConstants(expression) {
            const symbol = checker.getSymbolAtLocation(expression);
            if (symbol && shouldSymbolDefinitelyBeIgnoreInDeps(symbol)) {
                return true;
            }
            return isLiteralImmutable(expression.expression);
        }
        function isElementAccessConstants(expression) {
            const symbol = checker.getSymbolAtLocation(expression);
            if (symbol && shouldSymbolDefinitelyBeIgnoreInDeps(symbol)) {
                return true;
            }
            return (isLiteralImmutable(expression.expression) &&
                isExpressionConstants(expression.argumentExpression));
        }
        function isTemplateLiteralConstants(expression) {
            return expression.templateSpans.every(span => isExpressionConstants(span.expression));
        }
    }
    function shouldSymbolDefinitelyBeIgnoreInDeps(rawSymbol) {
        if (!cached.has(rawSymbol)) {
            cached.set(rawSymbol, peekSymbolDefinitelyBeIgnoreInDeps(rawSymbol));
        }
        return cached.get(rawSymbol);
    }
    function alreadyDuplicated(symbol) {
        return duplicated.has(symbol);
    }
    function markAsDuplicated(symbol) {
        duplicated.add(symbol);
    }
}
exports.createDepSymbolResolver = createDepSymbolResolver;
//# sourceMappingURL=resolver.js.map